;////////////////////////////////////////////////////////////////////////////////
;//
;//  ****************************************************************************
;//  * Project   : FASM DDK Test Drivers
;//  * Unit Name : crc.asm
;//  * Purpose   :  ,       
;//  *           :   -    
;//  * Author    :  (Rouse_) 
;//  * Copyright :  Fangorn Wizards Lab 1998 - 2009.
;//  * Version   : 1.00
;//  * Home Page : http://rouse.drkb.ru
;//  ****************************************************************************
;//

format PE native 4.0 at 10000h
entry DriverEntry

  include 'win32a.inc'
  include 'myincs.inc'
  include 'DDK\ntstatus.inc'
  include 'DDK\ntddk.inc'
  include 'DDK\stuff.inc'

;##########################################################################

section '.main' code readable writeable executable notpageable

;##########################################################################

	u_str	   usSymLink, '\??\CRCChecker'
	u_str	   usDevName, '\Device\CRCCheckerDevice'

	;dbgDriverEntry           db 'DriverEntry', 0
	;dbgDriverUnload          db 'DriverUnload', 0
	;dbgDispatchCreateClose   db 'DispatchCreateClose', 0
	;dbgDispatchControl       db 'DispatchControl', 0

;##########################################################################

	IOCTL_CALC_CRC = 22E000h

;##########################################################################

	FileStandardInformation = 5

	struct FILE_STANDARD_INFORMATION
	    AllocationSize LARGE_INTEGER <>
	    EndOfFile	   LARGE_INTEGER <>
	    NumberOfLinks  dd ?
	    DeletePending  db ?
	    Directory	   db ?
	    Dummee	   dw ?
	ends

;##########################################################################


;==========================================================================
proc DriverEntry pDriverObject, usRegistryPath
;==========================================================================
    locals
      status	    dd ?
      pDeviceObject dd ?
    endl
;==========================================================================

	;invoke     DbgPrint, dbgDriverEntry

	mov [status], STATUS_DEVICE_CONFIGURATION_ERROR

;-----[  

	lea	  eax, [pDeviceObject]
	invoke	  IoCreateDevice, [pDriverObject], 0, usDevName, FILE_DEVICE_UNKNOWN, 0, 0, eax
	or	  eax, eax
	mov	  [status], eax
	jnz	  DriverEntryExit

;-----[     ,       

	invoke	  IoCreateSymbolicLink, usSymLink, usDevName
	or	  eax, eax
	jz	  SetHandlers

;-----[     -  

	invoke	  IoDeleteDevice, [pDeviceObject]
	jmp	  DriverEntryExit

;-----[  

SetHandlers:
	mov	  eax, [pDriverObject]
	mov	  [eax + DRIVER_OBJECT.DriverUnload], DriverUnload
	mov	  [eax + DRIVER_OBJECT.MajorFunction + IRP_MJ_CREATE * 4], DispatchCreateClose
	mov	  [eax + DRIVER_OBJECT.MajorFunction + IRP_MJ_CLOSE * 4], DispatchCreateClose
	mov	  [eax + DRIVER_OBJECT.MajorFunction + IRP_MJ_DEVICE_CONTROL * 4], DispatchControl

	mov	  [status], STATUS_SUCCESS

DriverEntryExit:
	mov	  eax, [status]
	ret
endp

;==========================================================================
proc DriverUnload pDriverObject
;==========================================================================

	;invoke    DbgPrint, dbgDriverUnload

;-----[   

	invoke	  IoDeleteSymbolicLink, usSymLink

;-----[ ,   ?

	mov eax, [pDriverObject]
	mov eax, [eax + DRIVER_OBJECT.DeviceObject]
	or eax, eax
	jz AlreadyDeleted

;-----[     - 

	invoke	  IoDeleteDevice, eax

AlreadyDeleted:
	ret
endp

;==========================================================================
proc DispatchCreateClose pDeviceObject, pIrp
;==========================================================================

	;invoke     DbgPrint, dbgDispatchCreateClose

;-----[  / 

	mov	  eax, [pIrp]
	mov	  [eax + IRP.IoStatus.Status], STATUS_SUCCESS
	and	  [eax + IRP.IoStatus.Information], 0
	fastcall  IofCompleteRequest, [pIrp], IO_NO_INCREMENT
	mov	  eax, STATUS_SUCCESS
	ret
endp

;==========================================================================
proc DispatchControl pDeviceObject, pIrp
;==========================================================================
    locals
	status	  dd STATUS_INVALID_DEVICE_REQUEST
	ibl	  dd ?
	FilePath  UNICODE_STRING <>
	OA	  OBJECT_ATTRIBUTES <>
	hFile	  dd 0
	iosb	  IO_STATUS_BLOCK <>
	fsi	  FILE_STANDARD_INFORMATION <>
	Unicode   rb MAX_PATH * 2
	allocsize dd 0
	buff	  dd 0
	crcsize   dd 0
    endl
;==========================================================================

	;invoke    DbgPrint, dbgDispatchControl

	push	  esi edi

;-----[    IO_STACK_LOCATION

	mov	  edi, [pIrp]
	assume	  edi, IRP

	mov	  eax, [.edi.CurrentStackLocation]
	assume	  eax, IO_STACK_LOCATION

;-----[  IOCTL_ 

	mov	  ecx, [.eax.IoControlCode]
	cmp	  ecx, IOCTL_CALC_CRC
	jne	  Error

;-----[    ,     4 

	mov	  ecx, [.eax.OutputBufferLength]
	mov	  [status], STATUS_BUFFER_TOO_SMALL
	cmp	  ecx, 4
	jne	  Error

;-----[    ,    4   MAX_PATH - 1

	mov	  ecx, [.eax.InputBufferLength]
	mov	  [ibl], ecx
	mov	  [status], STATUS_INVALID_PARAMETER
	cmp	  ecx, 4
	jl	  Error
	cmp	  ecx, MAX_PATH - 1
	jg	  Error

;-----[  SEH 

	lea	  eax, [SehHandler]
	push	  eax
	push	  dword [fs:0]
	mov	  [fs:0], esp
	lea eax,  [SEHFinally]
	push eax
	push ebp

;-----[    Ansii  Uicode

	mov	  ecx, [ibl]
	mov	  esi, [.edi.SystemBuffer]
	lea	  edi, [Unicode]
copyloop:
	mov	  al,  byte [esi]
	mov	  byte [edi], al
	inc	  esi
	inc	  edi
	mov	  byte[edi], 0
	inc	  edi
	loop	  copyloop
	mov	  byte [edi], 0
	inc	  edi
	mov	  byte [edi], 0

;-----[    UNICODE_STRING

	mov	  eax, [ibl]
	shl	  eax, 1
	mov	  word [FilePath.Length], ax
	add	  eax, 2
	mov	  word [FilePath.MaximumLength], ax
	lea	  eax, [Unicode]
	mov	  [FilePath.Buffer], eax

;-----[    ( InitializeObjectAttributes)

	mov	  [OA.Length], sizeof.OBJECT_ATTRIBUTES
	mov	  [OA.RootDirectory], 0
	lea	  eax, [FilePath]
	mov	  [OA.ObjectName], eax
	mov	  [OA.Attributes], OBJ_CASE_INSENSITIVE + OBJ_KERNEL_HANDLE
	mov	  [OA.SecurityDescriptor], 0
	mov	  [OA.SecurityQualityOfService], 0

;-----[    

	mov	  eax, FILE_SYNCHRONOUS_IO_NONALERT
	push	  eax
	mov	  eax, FILE_SHARE_READ or FILE_SHARE_WRITE or FILE_SHARE_DELETE
	push	  eax
	lea	  eax, [iosb]
	push	  eax
	lea	  eax, [OA]
	push	  eax
	mov	  eax, FILE_READ_DATA or SYNCHRONIZE
	push	  eax
	lea	  eax, [hFile]
	push	  eax
	call	  [ZwOpenFile]

;-----[   

	mov	  [status], eax
	or	  eax, eax
	jnz	  Error

;-----[   

	lea	  eax, [iosb]
	lea	  ecx, [fsi]
	invoke	  ZwQueryInformationFile, [hFile], eax, ecx, \
	      sizeof.FILE_STANDARD_INFORMATION, FileStandardInformation

;-----[  

	mov	  [status], eax
	or	  eax, eax
	jnz	  CloseFile

;-----[     

	mov	  ecx, [fsi.EndOfFile.LowPart]
	mov	  [allocsize], ecx
	invoke	  ExAllocatePool, PagedPool, [allocsize]

;-----[ 

	mov	  [status], eax
	or	  eax, eax
	jz	  CloseFile

;-----[    

	mov	  [buff], eax
	invoke	  RtlZeroMemory, [buff], [allocsize]
	lea	  eax, [iosb]
	invoke	  ZwReadFile, [hFile], 0, NULL, NULL, eax, [buff], [allocsize], 0, NULL

;-----[ 

	mov	  [status], eax
	or	  eax, eax
	jnz	  FreeMem

;-----[    

	pushad
	mov	  esi, [buff]
	mov	  edi, [allocsize]
	cld
	xor	  ecx, ecx
	dec	  ecx
	mov	  edx, ecx
NextByteCRC:
	xor	  eax, eax
	xor	  ebx, ebx
	lodsb
	xor	  al, cl
	mov	  cl, ch
	mov	  ch, dl
	mov	  dl, dh
	mov	  dh, 8
NextBitCRC:
	shr	  bx, 1
	rcr	  ax, 1
	jnc	  NoCRC
	xor	  ax, 08320h
	xor	  bx, 0EDB8h
NoCRC:
	dec	  dh
	jnz	  NextBitCRC
	xor	  ecx, eax
	xor	  edx, ebx
	dec	  edi
	jnz	  NextByteCRC
	not	  edx
	not	  ecx
	mov	  eax, edx
	rol	  eax, 16
	mov	  ax, cx

;-----[       EAX,   

	mov	  [crcsize], 4
	mov	  [status], STATUS_SUCCESS
	mov	  edi, [pIrp]
	assume	  edi, IRP
	mov	  esi, [.edi.SystemBuffer]
	mov	  [esi], eax
	popad
	jmp	  FreeMem


;-----[  

SEHFinally:
	pop	  dword [fs:0]
	add	  esp, 4
	mov	  eax, STATUS_ACCESS_VIOLATION
	mov	  [status], eax
	mov	  ecx, [buff]
	or	  ecx, ecx
	jnz	  FreeMem
	mov	  ecx, [hFile]
	or	  ecx, ecx
	jz	  Error
	jmp	  CloseFile

;-----[  

FreeMem:
	invoke ExFreePool, [buff]

;-----[  

CloseFile:
	invoke	  ZwClose, [hFile]

Error:

;-----[  

	mov	  eax, [status]
	mov	  edi, [pIrp]
	mov	  [edi + IRP.IoStatus.Status], eax
	mov	  eax, [crcsize]
	mov	  [edi + IRP.IoStatus.Information], eax

DispatchControlExit:
	fastcall  IofCompleteRequest, [pIrp], IO_NO_INCREMENT

	pop	  edi esi
	mov	  eax, [status]
	ret
endp

;==========================================================================
proc SehHandler pExcept, pFrame, pContext, pDispatch
;==========================================================================
	push	  esi edi
	mov	  esi, [pFrame]
	mov	  edi, [pContext]
	mov	  eax, esi
	sub	  esi, 8
	add	  edi, $B4
	movsd			 ;   ebp
	movsd			 ;   eip
	mov	  [edi + 8], eax ;   esp
	pop	  edi esi
	xor	  eax, eax
	ret
endp


;##########################################################################

data import

    syslibrary ntoskrnl,'ntoskrnl.exe',\
	       hal,'hal.dll'

    include '\DDK\ntoskrnl.inc'
    include '\DDK\hal.inc'

end data

;##########################################################################

section '.reloc' data fixups readable discardable

;##########################################################################

